/* This is a driver to implement the linux API of /dev/js
 *
 * It was originally developed with a Logitech WingMan "Attack 2" joystick.
 * It may work with other USB joysticks -- it may not.
 * To find out, you have to adjust the USB identifier in the Makefile,
 * and see what happens.
 *
 * If you want in-depth examples for other USB type drivers for solaris,
 * download some of my other USB drivers  from
 *  http://www.bolthole.com/solaris/drivers/
 *
 *    Philip Brown,  October 2003
 *
 *  @(#) js.c 1.12@(#)
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>

/* Durnit, I want to use -N but dont know the exact magic */
char _depends_on[] = "misc/usba"; 

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/poll.h>
#include <sys/stream.h>
#include <stropts.h>

#include "joystick.h"

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


#define JS_MAX_TRANSFER 10 /* USB hardware setting */


/* make sure this is multiple of struct js_event size (currently 8 bytes)*/
#define JS_BUFFER_SIZE 512


/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"js"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	js_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	js_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	js_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	js_print(dev_t, char*);

static	int	js_chpoll(dev_t dev, short events, int anyyet,
                          short *reventsp, struct pollhead **phpp);

static	int	js_read(dev_t dev, struct uio* ustr, cred_t * credp);

static	int	js_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	js_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	js_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	js_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	js_cb_ops = {
	js_open,		/* open() */
	js_close,		/* close() */
	nodev,			/* strategy()		*/
	js_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	js_read,		/* read() */
	js_write,		/* write() */
	js_ioctl,			/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	js_chpoll,		/* chpoll() */
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP|D_64BIT,	/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops js_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	js_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	js_attach,		/* devo_attach */
	js_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&js_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"USB joystick driver",
	&js_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   js_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * js_glob_soft_handle;

/* This gets filled out on a per-instance basis by js_attach */
struct js_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex;
	kcondvar_t read_cv;
	
	int openflags; /* tell if we have been open()ed or not */

	struct pollhead pollhead; /* for chpoll() support */

	usb_config_descr_t usb_config;
	usb_endpoint_descr_t endpoint;
	int usb_interfacenum;
	usb_interface_descr_t usb_interface_d;

	usb_pipe_handle_t usb_pipe_handle;
	/* since we are not a streams driver we ahve to make our
	 * own buffer/queue
	 * "end" is where to add new data. "start" is where read()
	 * should take data from
	 */
	unsigned char *readbuffer;
	unsigned char *readbuf_start,*readbuf_end;

	 /* raw data is normally 4 bytes. But pad to 8, for future things*/
	uint8_t last_data[8];

	/* Joystick-hardware specific info here */
	/* At some point, we may be supporting multiple types */
	char *hw_name;
	int num_axes, num_buttons;

	/* Per-joystick axis adjustment values.
	 * First, scale up value with multiplier, to reach +-32767;
	 * Then subtract center-value, if axis values are positive-only
	 */
	int mult[6]; /* we only use 3 right now. but, plan ahead! */
	int center[6];

};



/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
js_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the js_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See js_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	cmn_err(CE_NOTE,"js: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&js_global_mutex, "js: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct js_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&js_glob_soft_handle,
				sizeof(struct js_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		cmn_err(CE_WARN,"soft_state_init failed");
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&js_global_mutex);
		ddi_soft_state_fini(&js_glob_soft_handle);
		cmn_err(CE_WARN, "js: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE,"js: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	js_debugprint("js: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	js_debugprint("js: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&js_glob_soft_handle);
		mutex_destroy(&js_global_mutex);
	}
	return (ret);
}

/************************************************************/

/**********************************************************************/
/*  Our read buffer handling section                                 */
/**********************************************************************/

/* Returns number of bytes in the read buffer.
 * Dont expect a sensible return unless you are holding the
 * instance_mutex.
 */
static int read_buffer_bytes(struct js_softstate *softstate){
	int datalen;


	if(softstate->readbuf_start > softstate->readbuf_end){
		datalen=(&softstate->readbuffer[JS_BUFFER_SIZE]) -
			softstate->readbuf_start;

		datalen+= softstate->readbuf_end - softstate->readbuffer;
	} else {
		datalen=softstate->readbuf_end - softstate->readbuf_start;
	}
	

	return datalen;
}

/*
 * Note that we have a circular buffer, and we do NOT overwrite.
 * After buffer is full, we drop further packets.
 * returns 0 on drop, 1 on normal
 *
 * YOU MUST HAVE instance_mutex BEFORE CALLING!!!
 */
static int
add_to_read_buffer(struct js_softstate *softstate, void *data, int len){
	unsigned char *bufstart, *bufend;
	unsigned char *bufdatastart,*bufdataend;
	int copylen, bytesfree;

	if(softstate==NULL) {
		js_debugprint("add_to_read_buffer: softstate==null");
		return 0;
	}
	if(data==NULL) {
		js_debugprint("add_to_read_buffer: data==null");
		return 1;
	}

	if(debug_level>1){
		js_debugprint("add_to_read_buffer called with len=%d\n",len);
	}
	
	bufstart=softstate->readbuffer;
	bufend=&softstate->readbuffer[JS_BUFFER_SIZE];
	bufdatastart=softstate->readbuf_start;
	bufdataend=softstate->readbuf_end;


	bytesfree=JS_BUFFER_SIZE - read_buffer_bytes(softstate);
	if(bytesfree==0){
		js_debugprint("add_to_read_buffer: buffer full\n");
		/* just discard */
		return 0;
	}

	if(len>bytesfree){
		/* This should never happen, since datasize should be
		* even multiple of buffer size
		*/
		js_debugprint("Note: add_to_read_buffer: partial overflow\n");
		return 0;
	} else {
		copylen=len;
	}

	if(&bufdataend[copylen]  >= &softstate->readbuffer[JS_BUFFER_SIZE]){
		copylen=(&softstate->readbuffer[JS_BUFFER_SIZE]) -
			softstate->readbuf_end;
	}
	bcopy(data,bufdataend, copylen);
	bufdataend += copylen;
	if(bufdataend==bufend){
		bufdataend=bufstart;
	}
	len-= copylen;
	if(len>0){
		/* should only be here if we had to wrap around end of buffer*/
		copylen=len;
		bcopy(data,bufdataend, copylen);
		bufdataend += copylen;
		len -= copylen;
	}
	
	softstate->readbuf_end=bufdataend;
	
	if(len!=0){
		cmn_err(CE_WARN, "js:add_to_read_buffer: internal error\n");
	}

	return 1;

} /* end of add_to_read_buffer() */

/* Make sure you have the instance_mutex before calling this.
 * Returns internal pointers to "clear" the buffer.
 */
static void read_buffer_reset(struct js_softstate *softstate){
	softstate->readbuf_start=softstate->readbuffer;
	softstate->readbuf_end=softstate->readbuffer;
}



/* for a given joystick type, translate the raw axis data
 * into the appropriate scale
 * The linux /dev/js driver is supposed to return values of
 *  -32768 to 32768, I believe.
 *
 */
static uint16_t
translate_axis(struct js_softstate *softstate, uint16_t data, int axisnum)
{

	data=data * softstate->mult[axisnum];
	return (data - softstate->center[axisnum]);
		
}


/* This takes raw data, and encodes it into "js_event" structs.
 * Right now, this is hard-wired to expect raw logitech wingman data
 *
 */
static void encode_js_data(struct js_softstate *softstate,uint8_t *data){
	struct js_event event;
	int bcount;
	uint8_t bmask;
	uint8_t lastbuttons=softstate->last_data[3];

	event.time=ddi_get_lbolt();

	if(data[0] !=softstate->last_data[0]){
		event.value=translate_axis(softstate,data[0],0);
		event.type=JS_EVENT_AXIS;
		event.number=0;
		if(add_to_read_buffer(softstate,&event,sizeof(event))){
			softstate->last_data[0] = data[0];
		}
	}
	if(data[1] !=softstate->last_data[1]){
		event.value=translate_axis(softstate,data[1], 1);
		event.type=JS_EVENT_AXIS;
		event.number=1;
		if(add_to_read_buffer(softstate,&event,sizeof(event))){
			softstate->last_data[1] = data[1];
		}

	}
	if(data[2] !=softstate->last_data[2]){
		event.value=translate_axis(softstate,data[2], 2);
		event.type=JS_EVENT_AXIS;
		event.number=2;
		if(add_to_read_buffer(softstate,&event,sizeof(event))){
			softstate->last_data[2] = data[2];
		}
	}
	if(data[3] == softstate->last_data[3]){
		return;
	}

	for(bcount=0; bcount <6; bcount++){
		bmask=1 <<bcount;
		if((data[3]&bmask )!= (lastbuttons&bmask)){
			event.value=(data[3]&bmask)? 1 : 0;
			event.type=JS_EVENT_BUTTON;
			event.number=bcount;
			add_to_read_buffer(softstate,&event,sizeof(event));
		}
	}
	softstate->last_data[3] = data[3];
	
}


static int
js_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	struct js_softstate *softstate=(struct js_softstate*)arg;
	unsigned char*chardata;
	int datalen;

	if(debug_level>1){
		js_debugprint("js_pipe_callback called\n");
	}

	if(mblkptr==NULL){
		printf("mblk==NULL\n");
		return (USB_SUCCESS);
	}

	mutex_enter(&softstate->instance_mutex);

	datalen=mblkptr->b_wptr - mblkptr->b_rptr;
	/*printf("theoretical len of data is %d\n",datalen);*/

	/* If there is joystick data, it is usually in 4-byte blocks,
	 * for the joystick we expect:
	 *  BYTE1 = x
	 *  BYTE2 = y
	 *  BYTE3 = z
	 *  BYTE4 = button-info
	 */

	chardata=mblkptr->b_rptr;

	if(datalen==4){
		encode_js_data(softstate,chardata);
	} else {
		js_debugprint("js_pipe_callback: discarding partial data\n");
	}

	freemsg(mblkptr);

	if(datalen==4){
		cv_signal(&softstate->read_cv);
	}
	mutex_exit(&softstate->instance_mutex);

	
	pollwakeup(&softstate->pollhead, POLLIN);

	if(debug_level>1){
		js_debugprint("js_pipe_callback: called pollwakeup\n");
	}

	return (USB_SUCCESS);
}

/* pretty much never called, but be paranoid */
static int
js_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
		      uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{

	js_debugprint("js_EXCEPTION_callback called: reason %d\n",
			complete_reason);
	return (USB_SUCCESS);
}


/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS to use.
 */
static int js_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int js_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct js_softstate *softstate;
	int returnval=DDI_FAILURE;
	int instance=getminor((dev_t)arg);
	
	js_debugprint("js: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(js_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* This routine can be called by js_attach (if errors occur)
 * and js_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void js_cleanup(struct js_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: js_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	ddi_soft_state_free(js_glob_soft_handle,instance);
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in js_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  js_detach()
 */
static int js_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct js_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;

	cmn_err(CE_NOTE,"js: attach called\n");

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	js_debugprint("js: js_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "js: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "js: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "js: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	js_debugprint("js: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(js_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "js: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(js_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "js: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(js_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	
	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		js_cleanup(softstate);
		return DDI_FAILURE;
	}


	js_debugprint("USB config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	js_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		js_debugprint("  (resetting interface num to 0)\n");
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		js_cleanup(softstate);
		return DDI_FAILURE;
	}

	
	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					0,/*endpoint index num*/
					&softstate->endpoint,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN, "could not parse endpoint 0");
		return DDI_FAILURE;
	}

		

	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);
	cv_init(&softstate->read_cv, NULL, CV_DRIVER,NULL);

	softstate->readbuffer=kmem_zalloc(JS_BUFFER_SIZE,KM_SLEEP);


	/* Now, at some time in the future, we may do some kind of
	 * fancy detection of what actual hardware we have.
	 * But right now, we are hardcoding for
	 * the "Logitech(tm) WingMan Attack 2(R)" USB joystick
	 */

	softstate->num_axes=3;
	softstate->num_buttons=6;
	{
		int x,y,z;

		x=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
		                   DDI_PROP_DONTPASS, "mult-0", 256);
		y=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
		                   DDI_PROP_DONTPASS, "mult-1", 256);
		z=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
		                   DDI_PROP_DONTPASS, "mult-2", 256);

		softstate->mult[0]=x;
		softstate->mult[1]=y;
		softstate->mult[2]=z;

		x=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
		                   DDI_PROP_DONTPASS, "center-0", 32760);
		y=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
		                   DDI_PROP_DONTPASS, "center-1", 32760);
		z=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
		                   DDI_PROP_DONTPASS, "center-2", 32760);

		softstate->center[0]=x;
		softstate->center[1]=y;
		softstate->center[2]=z;
		


		js_debugprint("mult=(%d,%d,%d)  center=(%d,%d,%d)\n",
			      softstate->mult[0],
			      softstate->mult[1],
			      softstate->mult[2],
			      softstate->center[0],
			      softstate->center[1],
			      softstate->center[2]);
	}

	softstate->hw_name="WingMan Attack 2";

	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);


	/* This little call all by itself, will put a whole bunch of
	 * useful information about the device in
	 * /var/adm/messages, aka "dmesg |tail"
	 */
	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

}

/* used by read() to figure out how many bytes of data
 * the user program wants right now
 */
static int count_uio_size(struct uio* uiostruct){
	int datalen=0;
	int vec_count;
	if(uiostruct==NULL){
		return 0;
	}
	for(vec_count=0;vec_count<uiostruct->uio_iovcnt;vec_count++){
		datalen+=uiostruct->uio_iov[vec_count].iov_len;
	}

	return datalen;
}


/* This support user calls to poll().
 * it probably also deals with O_NONBLOCK, but I dunno about that
 */
static int js_chpoll(dev_t dev, short eventmask, int anyyet,
                          short *reventsp, struct pollhead **phpp)
{
	int instance=DEVT_TO_INST(dev);
	struct js_softstate *softstate;
	short returnevents=0;

	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"js_chpoll failed: cannot get softstate?");
		return ENXIO;
	}

	if(debug_level>1){
		js_debugprint("chpoll() called\n");
	}

	if(softstate->readbuf_start != softstate->readbuf_end){
		returnevents|= POLLIN;
		if(debug_level>1){
			js_debugprint(" chpoll() will return POLLIN\n");
			}
	}
	/* in theory, should also check for whether to set POLLERR */

	if(!returnevents){
		/* this is just the "gotta do it" Solaris API mandate */
		if(!anyyet){
			*phpp = &softstate->pollhead;
		}
	}

	*reventsp = returnevents;
	return 0;
}



/* Note: user is expected to loop on reading (struct js_event)s
 * partial reads are an error of sorts. We will truncate to a
 * multiple of expected read size.
 * Note that there is a maximum allowed readahead size.
 *
 * We will 'block' on user read,  if not enough data is ready.
 * 
 */

/* return 0 on success, or errno */
static int js_read(dev_t dev, struct uio* ustr, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	int readbytes,copylen, nonblock=0;
	struct js_softstate *softstate;

	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"js_read failed: cannot get softstate?");
		return ENXIO;
	}

	readbytes=count_uio_size(ustr); /* how many bytes do they WANT?*/
	if(debug_level>1){
		js_debugprint("read(): user wants %d, we have %d\n",
		        readbytes,
			read_buffer_bytes(softstate));
	}
	
	if(ustr->uio_fmode & FNONBLOCK){
		if(debug_level>2){
			js_debugprint("read(): FNONBLOCK apparently set\n");
		}
		nonblock=1;
	}

	if(readbytes > JS_BUFFER_SIZE != 0){
		cmn_err(CE_WARN,
			"js_read: read size cannot be larger then %d\n",
			JS_BUFFER_SIZE);
		return EFAULT;/* is this the best errno to return? */
	}

	if(readbytes % sizeof(struct js_event) != 0){
		cmn_err(CE_WARN,
			"js_read: error size must be multiple of %d\n",
			sizeof(struct js_event));
		return EFAULT;/* is this the best errno to return? */

	}

	mutex_enter(&softstate->instance_mutex);

	/* Have to return whole read at once. Therefore, have to
	 * wait until read buffer fills up to required amount.
	 */
	while(read_buffer_bytes(softstate) <readbytes){
		int cstat;

		if(nonblock){
			mutex_exit(&softstate->instance_mutex);
			return (EAGAIN);
		}

		cstat=cv_wait_sig(&softstate->read_cv,
				      &softstate->instance_mutex);
		if(cstat<=0){
			js_debugprint("js_read: signal detected\n");
			mutex_exit(&softstate->instance_mutex);
			return (ENXIO);
		}
	}

	if(debug_level>1){
		js_debugprint("read(): after cv_wait, bytes in readbuff==%d\n",
			read_buffer_bytes(softstate));
	}

	/*******************************************************************/
	/* This section might be better off as a modular readbuffer routine*/
	/*******************************************************************/

	/* set copylen =  "how many bytes can we do in a single copy?" */
	if(softstate->readbuf_end < softstate->readbuf_start){
		copylen=(&softstate->readbuffer[JS_BUFFER_SIZE]) -
			softstate->readbuf_start;
	}else {
		copylen=softstate->readbuf_end - softstate->readbuf_start;
	}
	if(copylen>readbytes){
		copylen=readbytes;
	}

	if(uiomove(softstate->readbuf_start, copylen,UIO_READ,ustr)!=0){
		js_debugprint("js_read: uiomove failed?\n");
		mutex_exit(&softstate->instance_mutex);
		return EFAULT;
	}
	softstate->readbuf_start += copylen;
	if(softstate->readbuf_start ==
	   &softstate->readbuffer[JS_BUFFER_SIZE]) {
		softstate->readbuf_start=softstate->readbuffer;
	}
	if(softstate->readbuf_start >
	   &softstate->readbuffer[JS_BUFFER_SIZE]) {
		cmn_err(CE_WARN,"ERROR: js_read: internal bufptr error !!");
		softstate->readbuf_start=softstate->readbuffer;
	}

	readbytes -= copylen;

	if(readbytes!=0){
		/* If we are here, it is assumed that we had to wrap around
		 * the end of the buffer, and readbuf_start now points
		 * to the beginning of the whole buffer.
		 */

		copylen=readbytes;

		/* Although undocumented, empirical tests demonstrate that
		 * uiomove recognizes that the read request has been
		 * partially fulfilled, and when called again,
		 * does the second copy AFTER the previously copied data.
		 */
		if(uiomove(softstate->readbuf_start, copylen,UIO_READ,ustr)!=0){
			js_debugprint("js_read: uiomove2 failed?\n");
			mutex_exit(&softstate->instance_mutex);
			return EFAULT;
		}
		softstate->readbuf_start+=copylen;
	}
	/**************************************************/
	/*       End of readbuffer section                */
	/**************************************************/
	
	
	
	mutex_exit(&softstate->instance_mutex);
	return 0;
}

static int js_write(dev_t dev, struct uio *uiop, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct dfwheel_softstate *softstate;

   	js_debugprint("js: write called\n");

	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"js_write failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	return (ENXIO);
}


static int js_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct js_softstate *softstate;
	usb_pipe_policy_t policy,*policyptr;

	/* js_debugprint("open for instance %d\n", instance); */

	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

	if(openflags & (FNDELAY|FNONBLOCK)){
		js_debugprint("js_open: NONBLOCK flag detected\n");
	}

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=0){
		js_debugprint("js: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}

	policyptr=&policy;
	bzero(policyptr,sizeof(policy));

	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	policyptr->pp_callback_arg=(void*)softstate;/* same for ALL callbacks*/

	policyptr->pp_callback=js_pipe_callback;
	policyptr->pp_exception_callback=js_exception_callback;


	policyptr->pp_periodic_max_transfer_size=
		softstate->endpoint.wMaxPacketSize; /*JS_MAX_TRANSFER;*/

	read_buffer_reset(softstate);

	if(usb_pipe_open(softstate->dip,&softstate->endpoint, policyptr,
			 USB_FLAGS_SLEEP,
			 &softstate->usb_pipe_handle) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open failed");
		mutex_exit(&softstate->instance_mutex);
		return (ENXIO);
	}

	if(usb_pipe_start_polling(softstate->usb_pipe_handle,
				  USB_FLAGS_SLEEP| USB_FLAGS_SHORT_XFER_OK,
				  NULL,NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling of pipe1 FAILED");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}

	softstate->openflags=1;
	js_debugprint("DEBUG: start polling in js succeeded\n");
	mutex_exit(&softstate->instance_mutex);

   	return 0;
}

static int js_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct js_softstate *softstate;


	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"js_close failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	js_debugprint("js: close called\n");
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags==0){
		cmn_err(CE_WARN,"ERROR: js_close: pipe not open");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
		
	}

	usb_pipe_stop_polling(softstate->usb_pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);

	if(usb_pipe_close(&softstate->usb_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		js_debugprint("ERROR: usb_pipe_close failed?");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	}
	softstate->usb_pipe_handle=NULL;
	softstate->openflags=0;

	/* Okay, a little redundant since we do this in open(), but
	 * thats okay. It makes sure read_buffer_bytes returns 0.
	 */
	read_buffer_reset(softstate);
	
	mutex_exit(&softstate->instance_mutex);
   	return (0);

}

static int js_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp){

	int instance=DEVT_TO_INST(dev);
	uint32_t tmpint;
	uint8_t tmpchar;
	struct js_softstate *softstate;


	js_debugprint("ioctl called\n");

	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"js_ioctl failed: cannot get softstate?");
		return DDI_FAILURE;
	}
	if(cmd==I_NREAD){
		int numbytes;

		if(debug_level>1){
			js_debugprint("ioctl(I_NREAD) called\n");
		}

		numbytes=read_buffer_bytes(softstate);
		
		/* Okay we arent a STREAMS driver, but this is
		 * very convenient to support!
		 */
		if(arg==NULL){
			js_debugprint("I_NREAD passed NULL??\n");
			return EFAULT;
		}
		if(ddi_copyout(&numbytes,(int*)arg,sizeof(int),flags) != 0){
			js_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;
	}

	switch(cmd){
	    case JSIOCGVERSION:
		js_debugprint("ioctl(JSIOCGVERSION) called\n");
		tmpint=JS_VERSION;
		if(ddi_copyout(&tmpint,(int*)arg,sizeof(int),flags) != 0){
			js_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;

	    case JSIOCGAXES:
		tmpchar=softstate->num_axes;
		js_debugprint("ioctl(JSIOCGAXES) called. Returning %d\n",tmpchar);
		if(ddi_copyout(&tmpchar,(int*)arg,sizeof(tmpchar),flags) != 0){
			js_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;

	    case JSIOCGBUTTONS:
		tmpchar=softstate->num_buttons;
		js_debugprint("ioctl(JSIOCGBUTTONS) called. Returning %d\n",tmpchar);
		if(ddi_copyout(&tmpchar,(int*)arg,sizeof(tmpchar),flags) != 0){
			js_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;

	/*******************************************************/
	/* OLD linux 0.x js interface hacky stuff below here.  */
	/*  All "JS_XXX_XXX" stuff, I think                    */
	/*******************************************************/

	    case JS_GET_TIMELIMIT:
		tmpint=JS_DEF_TIMELIMIT;
		if(ddi_copyout(&tmpint,(int*)arg,sizeof(int),flags) != 0){
			js_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;
	    case JS_SET_TIMELIMIT:

		/* in theory, copyin the value and do 'something' with it.
		 * dunno what, though
		 */
		return 0;
	}

	tmpint = _IOC(_IOC_READ, 'j', 0x13, 0);
	if((cmd & tmpint) == tmpint){
		int namelen=strlen(softstate->hw_name)+1;
		js_debugprint("DEBUG: ioctl(JSIOCGNAME) called.Returning %s\n",
		              softstate->hw_name);
		tmpint=(cmd>>_IOC_SIZESHIFT) & _IOC_SIZEMASK;
		if(namelen>tmpint){
			namelen = tmpint;
		}
		
		if(ddi_copyout(softstate->hw_name,
		               (char*)arg, namelen, flags) != 0){
			js_debugprint("ioctl: ddi_copyout failed\n");
			return EFAULT;
		}
		return 0;
	}

	js_debugprint("do not recognize ioctl %d\n",cmd);

   	return (ENXIO);
}

static int js_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct js_softstate *softstate;
   	instance=ddi_get_instance(dip);

	js_debugprint("js: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(js_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"js_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	js_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}
